﻿namespace BMP32_Paint.NET_FileType
{
    //
    // 02/04/10, dev -- Be sure to add a reference to WindowsBase.
    //

    using System;
    using System.Drawing;
    using System.Drawing.Imaging;
    using System.IO;
    using System.Runtime.InteropServices;
    using System.Windows.Forms;
    using PaintDotNet;

    public class MyFileType : FileType
    {
        public MyFileType()
            : base(
                "32-bit Bitmap Image",
                FileTypeFlags.SupportsLoading | FileTypeFlags.SupportsSaving,
                new String[] { ".bmpx" })
        {
        }

        protected override Document OnLoad(Stream input)
        {
            try
            {
                byte[] imageData = new byte[54];
                input.Read(imageData, 0, 54);

                if (imageData[0] != (byte)'B' || imageData[1] != (byte)'M')
                {
                    throw new FormatException("Le fichier n'était pas un Bitmap correct");
                }

                Bitmap32FileHeader data = Bitmap32FileHeader.FromBytes(imageData);
                Bitmap b = null;
                if (data.BitCount != 32)
                {
                    // It's a normal bitmap
                    input.Position = 0;
                    b = new Bitmap(input);
                }
                else
                {
                    // It's a 32bit bitmap
                    int width = data.Width;
                    int height = data.Height;

                    input.Position = data.ImageOffset;

                    // Creation de l'image
                    b = new Bitmap(width, height, PixelFormat.Format32bppArgb);

                    for (int y = height - 1; y >= 0; y--)
                    {
                        for (int x = 0; x != width; x++)
                        {
                            // Lecture du pixel
                            byte[] colorData = new byte[4];
                            input.Read(colorData, 0, 4);

                            // Interprétation de la couleur
                            Color c = Color.FromArgb(
                                colorData[3],
                                colorData[2],
                                colorData[1],
                                colorData[0]);

                            // Modification de l'image d'origine
                            b.SetPixel(x, y, c);
                        }
                    }
                }

                return Document.FromImage(b);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                Bitmap b = new Bitmap(500, 500);
                return Document.FromImage(b);
            }
        }

        private struct Bitmap32FileHeader
        {
            public static Bitmap32FileHeader FromBytes(byte[] data)
            {
                Bitmap32FileHeader b = new Bitmap32FileHeader();

                b.FileSize = BitConverter.ToInt32(data, 2);
                b.Reserved = BitConverter.ToInt32(data, 6);
                b.ImageOffset = BitConverter.ToInt32(data, 10);
                b.InfoHeaderSize = BitConverter.ToInt32(data, 14);
                b.Width = BitConverter.ToInt32(data, 18);
                b.Height = BitConverter.ToInt32(data, 22);
                b.Planes = BitConverter.ToInt16(data, 26);
                b.BitCount = BitConverter.ToInt16(data, 28);
                b.Compression = BitConverter.ToInt32(data, 30);
                b.SizeImage = BitConverter.ToInt32(data, 34);
                b.HorizontalPixelsPerMeter = BitConverter.ToInt32(data, 38);
                b.VerticalPixelsPerMeter = BitConverter.ToInt32(data, 42);
                b.UsedColors = BitConverter.ToInt32(data, 46);
                b.ImportantColors = BitConverter.ToInt32(data, 50);

                return b;
            }

            public int FileSize;
            public int Reserved;
            public int ImageOffset;
            public int InfoHeaderSize;
            public int Width;
            public int Height;
            public short Planes;
            public short BitCount;
            public int Compression;
            public int SizeImage;
            public int HorizontalPixelsPerMeter;
            public int VerticalPixelsPerMeter;
            public int UsedColors;
            public int ImportantColors;
        }

        protected override void OnSave(
            Document input,
            Stream output,
            SaveConfigToken token,
            Surface scratchSurface,
            ProgressEventHandler callback)
        {
            const float InchesPerMeter = 39.3700787f;
            RenderArgs ra = new RenderArgs(new Surface(input.Size));
            
            //
            // 02/04/10, dev -- Render(RenderArgs args) deprecated.  Added "true" parameter.
            //
            input.Render(ra, true);

            Bitmap bitmap = ra.Bitmap;
            byte[] bytesFor0 = BitConverter.GetBytes(0);
            byte[] bytesFor1 = BitConverter.GetBytes(1);

            const int reserved = 0;
            
            //
            // 02/04/10, dev -- Changed from 108 to 40.  Our offset will always be 0x28 bytes.
            //
            const int infoSize = 40;
            const int headerSize = infoSize + 14;
            int width = bitmap.Width;
            int height = bitmap.Height;
            int imageSize = width * height * 4;
            int fileSize = headerSize + imageSize;
            int horizontalPixelsPerMeter = (int)Math.Round(bitmap.HorizontalResolution * InchesPerMeter, 0);
            int verticalPixelsPerMeter = (int)Math.Round(bitmap.VerticalResolution * InchesPerMeter, 0);
            output.WriteByte((byte)'B');
            output.WriteByte((byte)'M');
            output.Write(BitConverter.GetBytes(fileSize), 0, 4);
            output.Write(BitConverter.GetBytes(reserved), 0, 4);
            output.Write(BitConverter.GetBytes(headerSize), 0, 4);
            output.Write(BitConverter.GetBytes(infoSize), 0, 4);
            output.Write(BitConverter.GetBytes(width), 0, 4);
            output.Write(BitConverter.GetBytes(height), 0, 4);
            output.Write(BitConverter.GetBytes((short)1), 0, 2);
            output.Write(BitConverter.GetBytes((short)32), 0, 2);

            //
            // 02/04/10, dev -- Changed from 3 (Bit Field) to 0 (No compression).
            //
            output.Write(BitConverter.GetBytes(0), 0, 4); // Compression

            output.Write(BitConverter.GetBytes(imageSize), 0, 4);
            output.Write(BitConverter.GetBytes(horizontalPixelsPerMeter), 0, 4);
            output.Write(BitConverter.GetBytes(verticalPixelsPerMeter), 0, 4);
            output.Write(bytesFor0, 0, 4); // ClrUsed
            output.Write(bytesFor0, 0, 4); // ClrImportant
            
            //
            // 02/04/10, dev -- Removed section.
            //

            if (bitmap.PixelFormat == PixelFormat.Format32bppArgb)
            {
                BitmapData data = null;
                try
                {
                    data = bitmap.LockBits(
                        new Rectangle(0, 0, width, height),
                        ImageLockMode.ReadOnly,
                        PixelFormat.Format32bppArgb);
                    int stride = data.Stride;
                    byte[] line = new byte[Math.Abs(stride)];
                    if (stride > 0)
                    {
                        IntPtr ptr = new IntPtr((long)data.Scan0 + (height - 1) * stride);
                        for (int i = 0; i < height; i++)
                        {
                            Marshal.Copy(ptr, line, 0, line.Length);
                            output.Write(line, 0, line.Length);
                            ptr = new IntPtr((long)ptr - stride);
                        }
                    }
                    else
                    {
                        IntPtr ptr = data.Scan0;
                        for (int i = 0; i < height; i++)
                        {
                            Marshal.Copy(ptr, line, 0, line.Length);
                            output.Write(line, 0, line.Length);
                            ptr = new IntPtr((long)ptr + stride);
                        }
                    }
                }
                finally
                {
                    if (data != null)
                    {
                        bitmap.UnlockBits(data);
                    }
                }
            }
            else
            {
                for (int y = height - 1; y >= 0; y--)
                {
                    for (int x = 0; x < width; x++)
                    {
                        Color pixel = bitmap.GetPixel(x, y);
                        output.WriteByte(pixel.B);
                        output.WriteByte(pixel.G);
                        output.WriteByte(pixel.R);
                        output.WriteByte(pixel.A);
                    }

                    callback(this, new ProgressEventArgs(100 * (height - y) / height));
                }
            }    
        }
    }

    public class MyFileTypeFactory : IFileTypeFactory
    {
        public FileType[] GetFileTypeInstances()
        {
            return new FileType[] { new MyFileType() };
        }
    }
}
